# pylint: disable=W0201, C0103
""" A class to provide authentication via AWS Cognito """
import base64
import logging
import urllib
from tornado.web import RequestHandler, HTTPError
from tornado.auth import OAuth2Mixin
from tornado import escape
from bluemax.web.authentication import UserMixin

LOGGER = logging.getLogger(__name__)


class CognitoHandler(UserMixin, RequestHandler, OAuth2Mixin):  # pylint: disable=W0223
    """
        This plays safe with AWS Congnito
        Settings should have:

        "cookie_name": "tornado-user",
        "cookie_secret": "it_was_a_dark_and_stormy_night_cognito",
        "login_url": "https://<provided by ngrok>.ngrok.io/login",
        "cognito_oauth": {
            "key": "<client_id provided by cognito>",
            "secret": "<secret provided by cognito>",
            "redirect_url": "https://<provided by ngrok>.ngrok.io/login",
            "endpoint": "https://<provided by cognito>.amazoncognito.com",
        }

    """

    def initialize(self):
        """ This is how tornado recommends - obviously not lint """
        self._OAUTH_SETTINGS_KEY = "cognito_oauth"
        cognito_domain = self.settings[self._OAUTH_SETTINGS_KEY]["endpoint"]
        self._OAUTH_AUTHORIZE_URL = f"{cognito_domain}/oauth2/authorize"
        self._OAUTH_ACCESS_TOKEN_URL = f"{cognito_domain}/oauth2/token"
        self._OAUTH_USER_INFO_URL = f"{cognito_domain}/oauth2/userInfo"
        self._OAUTH_LOGIN_URL = f"{cognito_domain}/oauth2/login"
        self._REDIRECT_URL = self.settings[self._OAUTH_SETTINGS_KEY]["redirect_url"]

    async def get(self):
        """ Here we go: called, then called back! """
        code = self.get_argument("code", False)
        if code:
            LOGGER.debug("have code")
            http = self.get_auth_http_client()
            access = await self.get_authenticated_user(
                redirect_uri=self._REDIRECT_URL, code=code
            )
            LOGGER.debug("have access %r", access)
            response = await http.fetch(
                self._OAUTH_USER_INFO_URL,
                headers={"Authorization": f"Bearer {access['access_token']}"},
                raise_error=False,
            )
            if response.code != 200:
                LOGGER.debug("failed: %s %r", response.code, response.body)
                raise HTTPError(response.code)
            user = escape.json_decode(response.body)
            LOGGER.debug("have user: %r", user)
            # Save the user and access token with
            # e.g. set_secure_cookie.
            self.set_current_user(user)
            self.redirect("/")
        else:
            try:
                LOGGER.debug("redirecting")
                await self.authorize_redirect(
                    redirect_uri=self._REDIRECT_URL,
                    client_id=self.settings["cognito_oauth"]["key"],
                    scope=["profile", "email", "openid"],
                    response_type="code",
                    extra_params={},
                )
                LOGGER.debug("redirected")
            except Exception as ex:
                LOGGER.exception(ex)
                raise

    async def get_authenticated_user(self, redirect_uri: str, code: str):
        """ Ask cognito """
        client_id = self.settings["cognito_oauth"]["key"]
        client_secret = self.settings["cognito_oauth"]["secret"]
        http = self.get_auth_http_client()
        body = urllib.parse.urlencode(
            {
                "redirect_uri": redirect_uri,
                "code": code,
                "client_id": client_id,
                "grant_type": "authorization_code",
            }
        )
        authorized = base64.b64encode(
            f"{client_id}:{client_secret}".encode("utf-8")
        ).decode("ascii")
        LOGGER.debug("basic %s", authorized)
        response = await http.fetch(
            self._OAUTH_ACCESS_TOKEN_URL,
            method="POST",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Authorization": f"Basic {authorized}",
            },
            body=body,
        )
        return escape.json_decode(response.body)



"""
    Support for Tornado Authentication.
"""
import logging
from bluemax.web import json_utils

LOGGER = logging.getLogger(__name__)


class UserMixin:
    """
        Should be mixed in with WebsocketHandlers
        or RequestHandlers. Make sure it is the
        first class in the declaration as it
        overrides methods in both handlers.
    """

    @property
    def cookie_name(self):
        """ return the cookie_name declared in options"""
        return self.settings.get("cookie_name")

    def get_current_user(self):
        """ return the current user from the cookie """
        result = self.get_secure_cookie(self.cookie_name)
        LOGGER.debug("%s:%s", self.cookie_name, result)
        if result:
            result = json_utils.loads(result.decode("utf-8"))
        return result

    def set_current_user(self, value):
        """ put the current user in the cookie """
        if value:
            self.current_user = value
            self.set_secure_cookie(self.cookie_name, json_utils.dumps(value))
        else:
            self.clear_cookie(self.cookie_name)



"""
    This is a our invoke program
"""
import os
import sys
import logging
from invoke import Program, Collection
import bluemax
from . import create
from . import run

sys.path.insert(0, os.getcwd())

logging.basicConfig(level=logging.DEBUG)

_NAMESPACE_ = Collection()
try:
    from bluemax.sa import database

    _NAMESPACE_.add_collection(Collection.from_module(database))
except ImportError:
    pass
_NAMESPACE_.add_collection(Collection.from_module(create))
_NAMESPACE_.add_collection(Collection.from_module(run))

program = Program(
    version=bluemax.VERSION, namespace=_NAMESPACE_
)  # pylint: disable=C0103



# pylint: disable=W0703,R1710
""" Useful with dates """
import datetime


def json_date(date):
    """ javascript hates a t """
    if date:
        return date.isoformat().replace("T", " ")


def parse_date(value):  # noqa: 901
    """
        Returns a Python datetime.datetime object,
        the input must be in some date ISO format
    """
    if isinstance(value, (datetime.date, datetime.datetime)):
        return value
    result = None
    if value:
        try:
            result = datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")
        except Exception:
            try:
                result = datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S.%f")
            except Exception:
                try:
                    result = datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
                except Exception:
                    result = datetime.datetime.strptime(value, "%Y-%m-%d")
    return result


def parse_unix_time(value):
    """ can never remeber how to get back from javascript """
    return datetime.datetime.fromtimestamp(float(value))



import asyncio
import importlib
import logging
from bluemax.rpc import ContextRpc
from .services_manager import ServicesManager

LOGGER = logging.getLogger(__name__)


class ServicesRpc(ServicesManager, ContextRpc):
    def __init__(self, package, services, *args, **kwargs):
        ServicesManager.__init__(self, importlib.import_module(services))
        ContextRpc.__init__(self, package, *args, **kwargs)
        asyncio.ensure_future(self.start_all_services())

    @property
    def procedures(self):
        """ return a list of publicly available functions """
        return getattr(self._package_, "__all__") + self._SERVICE_ACTIONS_

    def qname_to_func(self, qname):
        """ return the service or function for the qname """
        if qname in self._SERVICE_ACTIONS_:
            return getattr(self, qname)
        return getattr(self._package_, qname)

    async def shutdown(self, wait=True):
        """ We have runnning resources that require tidy up """
        await self.stop_all_services()
        await ContextRpc.shutdown(self)



"""
    Support for Tornado Authentication.
"""
from tornado.web import RequestHandler
from .user_mixin import UserMixin


class LogoutHandler(UserMixin, RequestHandler):  # pylint: disable=W0223
    """ We need to logout """

    def get(self):
        """ removes cookie and redirects to optional next argument """
        self.set_current_user(None)
        self.redirect(self.get_argument("next", "/"))



"""
    We wanted something like Thread and Process Pool Exceutors
    but for async function
"""
import asyncio
import logging
from concurrent.futures import Executor

LOGGER = logging.getLogger(__name__)


class NotRunningException(Exception):
    """ Called when a running pool is required """


class AsyncPoolExecutor(Executor):
    """
        This object can be called with async functions and you
        can limit the number of workers with max_workers.
        It maintains a queue of work of max_size.
    """

    def __init__(self, loop=None, max_workers=None, max_size=0):
        self.max_workers = max_workers if max_workers else 4
        self.running = False
        self.loop = loop if loop else asyncio.get_event_loop()
        self.queue = asyncio.Queue(maxsize=max_size, loop=self.loop)
        self.workers = []

    def start(self):
        """ called by aenter """
        LOGGER.info("starting...")
        self.running = True
        for n_id in range(self.max_workers):
            worker = asyncio.ensure_future(self.worker(n_id))
            self.workers.append(worker)

    async def worker(self, n_id):
        """ worker called a task """
        LOGGER.debug("working: %s", n_id)
        while self.running is True:
            try:
                await self.perform(n_id)
            except asyncio.CancelledError:
                return

    async def perform(self, n_id):
        """ called by a worker to get queue item """
        LOGGER.debug("pulling from queue: %s", n_id)
        item = await self.queue.get()
        results, func, args, kwargs = item
        try:
            result = await func(*args, **kwargs)
            results.set_result(result)
        except Exception as ex:  # pylint: disable=W0703
            results.set_exception(ex)
        finally:
            self.queue.task_done()

    def submit(self, fn, *args, **kwargs):
        """ part of the executor api """
        if self.running is False:
            raise NotRunningException()
        result = self.loop.create_future()
        LOGGER.info("adding to queue...")
        self.queue.put_nowait((result, fn, args, kwargs))
        return result

    def map(self, func, *iterables):  # pylint: disable=W0221
        """ part of the executor api """
        if self.running is False:
            raise NotRunningException()
        results = []
        args = zip(*iterables)
        LOGGER.info("args... %r", args)
        for line in args:
            results.append(self.submit(func, *line))
        return results

    async def shutdown(self, wait=True):
        """
            part of the executor api
            called by aexit
        """
        LOGGER.info("shutting down...")
        if wait:
            LOGGER.info("joining...")
            await self.queue.join()
        self.running = False
        try:
            while True:
                LOGGER.info("cancelling queue contents...")
                item = self.queue.get_nowait()
                item[0].cancel()
                self.queue.task_done()
        except asyncio.QueueEmpty:
            pass
        LOGGER.info("cancelling workers...")
        for task in self.workers:
            task.cancel()
        LOGGER.info("gathering workers...")
        await asyncio.gather(*self.workers, return_exceptions=True)
        self.workers.clear()
        LOGGER.info("shutdown")

    async def __aenter__(self):
        """ This makes us an async contextmanager """
        self.start()
        return self

    async def __aexit__(self, *args):
        """ This is called when the with is up """
        await self.shutdown()



"""
    Web server with rpc handler
"""
import logging
import importlib
import tornado.web
from tornado.ioloop import IOLoop
from tornado.web import StaticFileHandler
from bluemax.settings import get_settings, extend
from .broadcaster import Broadcaster
from .authentication import LogoutHandler
from .auth_static_file_handler import AuthStaticFileHandler
from .rpc_websocket import RpcWebsocket, RpcHandler

BROADCASTER = Broadcaster()


def qname_to_class(qname):
    """ returns a class from module_name:class_name """
    _modules, _class = qname.split(":")
    return getattr(importlib.import_module(_modules), _class)


def make_app():
    """ Return a tornado app with settings and routes """
    config = get_settings("tornado", {})
    assert get_settings("procedures"), "procedures required to remote!"
    manager = qname_to_class(config.get("MANAGER", "bluemax.rpc:ContextRpc"))
    settings = {
        "debug": config.get("DEBUG", get_settings("DEBUG", False)),
        "manager": manager.rpc_for(
            get_settings("procedures"), services=get_settings("services")
        ),
        "cookie_name": config.get("COOKIE_NAME", "blueq-user"),
        "cookie_secret": config.get(
            "COOKIE_SECRET", "<change me - it was a dark and stormy..>"
        ),
    }
    routes = [(r"/rpc", RpcHandler), (r"/ws", RpcWebsocket)]
    if config.get("AUTH_HANDLER"):
        auth_handler = qname_to_class(config.get("AUTH_HANDLER"))
        settings["login_url"] = "/login"
        routes.insert(0, (r"/logout", LogoutHandler))
        routes.insert(0, (r"/login", auth_handler))
        if config.get("STATIC_DIR"):
            routes.append(
                (
                    r"/(.*)",
                    AuthStaticFileHandler,
                    {
                        "path": config.get("STATIC_DIR"),
                        "default_filename": "index.html",
                    },
                )
            )
            logging.info("serving from %s", config.get("STATIC_DIR"))
    elif config.get("STATIC_DIR"):
        routes.append(
            (
                r"/(.*)",
                StaticFileHandler,
                {"path": config.get("STATIC_DIR"), "default_filename": "index.html"},
            )
        )
        logging.info("serving from %s", config.get("STATIC_DIR"))
    if get_settings("urls_extend"):
        extend(get_settings("urls_extend"), routes)
    app = tornado.web.Application(routes, **settings)
    BROADCASTER.start(settings["manager"].get_broadcast_queue(), RpcWebsocket.broadcast)
    return app


def main():
    """ run the tornado web server """
    app = make_app()
    port = int(get_settings("PORT", "8080"))
    app.listen(port)
    logging.info("listening on port %s", port)
    if app.settings.get("debug") is True:
        logging.info("running in debug mode")
    ioloop = IOLoop.current()
    try:
        ioloop.start()
    except KeyboardInterrupt:
        logging.info("stopping")
        ioloop.stop()



"""
    We want to provide a remote procedure context
"""
import asyncio
import functools
import importlib
import logging
from concurrent.futures import ThreadPoolExecutor
from .asyncpoolexecutor import AsyncPoolExecutor
from . import context


LOGGER = logging.getLogger(__name__)


class Rpc:
    """ Base class for rpcs """

    def __init__(self, package, *args, **kwargs):
        """ Given a package """
        self._package_ = package
        LOGGER.info("rpc package available: %s", self._package_)

    @property
    def procedures(self):
        """ return a list of publicly available functions """
        return getattr(self._package_, "__all__")

    def perform(self, qname, *args, **kwargs):
        """ perform the function and result the result """
        LOGGER.debug("perform: %s", qname)
        func = getattr(self._package_, qname)
        result = func(*args, **kwargs)
        return result

    def broadcast(self, signal, message, client_filter=None):
        """ subclass responsibly: noop! """

    def broadcast_later(self, messages):
        """ subclass responsibly: noop! """

    @classmethod
    def rpc_for(cls, module_name, *args, **kwargs):
        """ will import a module_name and return an rpc """
        LOGGER.info("loading module: %s", module_name)
        package = importlib.import_module(module_name)
        return cls(package, *args, **kwargs)


class AsyncRpc(Rpc):
    """ We need to perform async function """

    async def perform(self, qname, *args, **kwargs):
        """ If the result is a coro - await it """
        result = super().perform(qname, *args, **kwargs)
        if asyncio.iscoroutine(result):
            LOGGER.debug("awaiting: %s", qname)
            result = await result
        return result


class PoolRpc(AsyncRpc):
    """
        We can restrict the number of workers
    """

    def __init__(self, *args, loop=None, max_workers=None, max_threads=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._loop = loop if loop else asyncio.get_event_loop()
        self._thread_pool_ = ThreadPoolExecutor(max_workers=max_threads)
        self._async_pool_ = AsyncPoolExecutor(loop=self._loop, max_workers=max_workers)
        self._async_pool_.start()

    def perform(self, qname, *args, **kwargs):
        """ Submit func to the right pool """
        func = getattr(self._package_, qname)
        if asyncio.iscoroutinefunction(func):
            pool = self._async_pool_
        else:
            pool = self._thread_pool_
        proc = functools.partial(func, *args, **kwargs)
        future = self._loop.run_in_executor(pool, proc)
        return future

    async def shutdown(self, wait=True):
        """ We have runnning resources that require tidy up """
        self._thread_pool_.shutdown(wait=wait)
        await self._async_pool_.shutdown(wait=wait)
        LOGGER.info("shutdown")


class BroadcastMixin:
    """ How we get the message out """

    def get_broadcast_queue(self):
        """ returns the queue containing signals for clients """
        return self._broadcast_q_

    def _broadcast_(self, signal, message, client_filter=None):
        """ Add message to broadcast queue """
        LOGGER.debug("queuing broadcast: %s", message)
        self._broadcast_q_.put_nowait((signal, message, client_filter))

    def broadcast(self, signal, message, client_filter=None):
        """ Add message to broadcast queue thread safe"""
        self._loop.call_soon_threadsafe(
            self._broadcast_, signal, message, client_filter
        )

    def broadcast_later(self, messages):
        """ Will add messages from any thread back to our _loop """
        for message in messages:
            self._loop.call_soon_threadsafe(self._broadcast_, *message)


class ContextRpc(BroadcastMixin, PoolRpc):
    """
        We execute in a context
    """

    def __init__(self, *args, **kwargs):
        """ introduce a broadcast queue """
        context._MANAGER_ = self  # pylint: disable=W0212
        self._broadcast_q_ = asyncio.Queue()
        super().__init__(*args, **kwargs)

    def qname_to_func(self, qname):
        """ return the function for the qname """
        return getattr(self._package_, qname)

    def perform(self, user, qname, *args, **kwargs):  # pylint: disable=W0221
        """ Submit func to the right pool and context """

        func = self.qname_to_func(qname)
        logging.debug("calling %s", func)
        if asyncio.iscoroutinefunction(func):
            pool = self._async_pool_
            proc = functools.partial(context.acontext_call, func, user, args, kwargs)
        else:
            pool = self._thread_pool_
            proc = functools.partial(context.context_call, func, user, args, kwargs)
        future = self._loop.run_in_executor(pool, proc)
        return future



""" A manager that delegates to redis """
import asyncio
import logging
from bluemax.settings import get_settings
from bluemax.rpc import BroadcastMixin, Rpc
from bluemax.web import json_utils
from bluemax.performer import Performer
from bluemax.services import ServicesManager
from .redis_pubsub import redis_subscribe
from . import rd_session

LOGGER = logging.getLogger(__name__)


class RemoteException(Exception):
    """ We cannot serialise Exceptions so this is the message """


class RedisManager(BroadcastMixin, Performer, Rpc):
    """
        A subclass of manager that delegates
        to a redis queue. It also subscribes
        to broadcasts and sends them on.
    """

    def __init__(self, *args, **kwargs):
        """ Store state and schedule setup """
        BroadcastMixin.__init__(self)
        Performer.__init__(self)
        Rpc.__init__(self, *args, **kwargs)
        self._broadcast_q_ = asyncio.Queue()
        self._loop = asyncio.get_event_loop()
        self._futures_ = {}
        # set off subscriptions
        self.handle_broadcast()

    async def shutdown(self):
        """ Shut down the pool """
        redis = rd_session.pool()
        redis.close()

    @property
    def procedures(self):
        """ return a list of publicly available functions """
        procs = getattr(self._package_, "__all__")
        if get_settings("services"):
            procs = procs + ServicesManager._SERVICE_ACTIONS_
        return procs

    @redis_subscribe(key=get_settings("redis_broadcast_q", "broadcast"))
    def handle_broadcast(self, message=None):
        """ called by redis sub """
        LOGGER.info("got broadcast: %s", message)
        args = json_utils.loads(message)
        self._broadcast_(*args)



from invoke import task
from . import create


@task
def sidney(_, name):
    """ all the jazz - creates everything."""
    from bluemax.sa import database

    create.project(_, name, with_config=True)
    create.tests(_, name)
    database.create(_, name)



import asyncio
import logging
import uuid
from bluemax.web import json_utils
from bluemax.settings import get_settings
from bluemax.work.redis_pubsub import redis_publish, redis_subscribe
from bluemax.services import ServicesManager


LOGGER = logging.getLogger(__name__)


class RemoteException(Exception):
    """ We cannot serialise Exceptions so this is the message """


class Performer:
    def __init__(self):
        self._futures_ = {}
        self.REPY_CHANNEL = f"channel:{uuid.uuid4()}"
        redis_subscribe(self.handle_reply, key=self.REPY_CHANNEL)()

    @redis_publish(key=get_settings("redis_broadcast_q", "broadcast"))
    def broadcast(self, signal, message, filter_clients=None):  # pylint: disable=W0221
        """ broadcast, because there may be more than one """
        return json_utils.dumps((signal, message, filter_clients))

    @redis_publish(key=get_settings("redis_work_q", "perform"), cmd="rpush")
    def send_perform(self, content):  # pylint: disable=R0201
        """ format performance content """
        return json_utils.dumps(content)

    @redis_publish(key=get_settings("redis_service_q", "service"))
    def send_service(self, content):  # pylint: disable=R0201
        """ format performance content """
        return json_utils.dumps(content)

    def perform(self, user, qname, *args, **kwargs):  # pylint: disable=W0221
        """ create task add it to queue """
        future_id = str(uuid.uuid4())
        content = {
            "user": user,
            "qname": qname,
            "args": args,
            "kwargs": kwargs,
            "future_id": future_id,
            "reply": self.REPY_CHANNEL,
        }
        future = None
        LOGGER.info(content)
        if qname in ServicesManager._SERVICE_ACTIONS_:
            self.send_service(content)
        else:
            future = asyncio.Future()
            self._futures_[future_id] = future
            self.send_perform(content)
        return future

    def handle_reply(self, msg=None):
        """ called by redis sub """
        result = json_utils.loads(msg)
        LOGGER.debug("response: %s", result)
        future = self._futures_.get(result.get("future_id"), None)
        if future:
            if result.get("result"):
                future.set_result(result["result"])
            else:
                future.set_exception(RemoteException(result.get("error")))
        if result.get("broadcasts"):
            for message in result["broadcasts"]:
                self.broadcast(*message)



"""
    Scripts to make alembic simple
"""
import os
import shutil
from configparser import SafeConfigParser
from alembic.config import Config
from invoke import task
from bluemax.tasks.utils import confirm_action


def confirm_db_action(action, color="0;31"):
    """db are you sure? """
    db_url = get_db_url()
    return confirm_action(f"\033[{color}m{action}\033[0m {db_url}")


@task
def create_model(_, project):
    """ creates a sqlalchemy model package inside project """
    model_path = os.path.abspath(os.path.join(project, "model"))
    if os.path.isdir(model_path):
        print("model exists")
        return
    os.makedirs(model_path)
    open(os.path.join(model_path, "__init__.py"), "w")
    with open(os.path.join(model_path, "base.py"), "w") as file:
        file.write(
            '''""" Used by alembic """
from sqlalchemy.ext.declarative import declarative_base, declared_attr


class Base:
    """ Base for orm model classes """

    @declared_attr
    def __tablename__(cls):
        """ with this tablename is class name lower"""
        return cls.__name__.lower()


Base = declarative_base(cls=Base)
'''
        )


@task(default=True)
def create(ctx, name, project=None):
    """ Creates an alembic managed db """
    if project is None:
        project = name
    if not os.path.isdir("database"):
        os.makedirs("database")

    ctx.run(f"alembic init database/{name}")
    parser = SafeConfigParser()
    with open("alembic.ini", "r") as configfile:
        parser.read_file(configfile)
        for key, value in parser.items():
            print(key, value)
    parser.set("alembic", "sqlalchemy.url", f"sqlite:///{name}.sqlite3")
    with open("alembic.ini", "w") as configfile:
        parser.write(configfile)
    with open(os.path.join("database", name, "env.py"), "r") as file:
        contents = file.read()
        contents = contents.split("target_metadata = None\n")

    with open(os.path.join("database", name, "env.py"), "w") as file:
        file.write(contents[0])
        file.write("\n")
        file.write(f"from {project}.model.base import Base\n")
        file.write("target_metadata = Base.metadata\n")
        file.write(contents[1])

    create_model(ctx, project)


@task
def remove(_, name):
    """ removes a database of name from the database folder """
    db_path = os.path.abspath(os.path.join("database", name))
    if os.path.isdir(db_path):
        shutil.rmtree(db_path)
        print(f"removed {db_path}")
    alembic_file = "alembic.ini"
    if os.path.isfile(alembic_file):
        os.unlink(alembic_file)
        print(f"removed {alembic_file}")


def _run_with_pythonpath(ctx, cmd):
    """ extend the python path environment variable with cwd """
    path = os.getenv("PYTHONPATH", "")
    env = {"PYTHONPATH": ".:" + path}
    return ctx.run(cmd, env=env)


def get_db_url():
    """ get the db_url from alembic.ini """
    alembic_cfg = Config("alembic.ini")
    return alembic_cfg.get_main_option("sqlalchemy.url")


@task
def auto_revision(ctx, message):
    """ creates an autogenerated revision in alembic """
    _run_with_pythonpath(ctx, f'alembic revision --autogenerate -m "{message}"')


@task
def upgrade(ctx, revision="head"):
    """ upgrades with alembic to revision """
    if confirm_db_action("UPGRADE", color="0;32"):
        _run_with_pythonpath(ctx, f"alembic upgrade {revision}")
    # with engine.begin() as connection:
    #     alembic_cfg.attributes['connection'] = connection
    #     command.upgrade(alembic_cfg, "head")


@task
def downgrade(ctx, revision):
    """ downgrades to revsion with alembic - base is nothing!"""
    if confirm_db_action("DOWNGRADE"):
        _run_with_pythonpath(ctx, f"alembic downgrade {revision}")



"""
    Support for Tornado Authentication.
"""
import asyncio
from tornado.web import RequestHandler, HTTPError
from .user_mixin import UserMixin


class LoginHandler(UserMixin, RequestHandler):  # pylint: disable=W0223
    """
        Can be called as ajax from the
        websocket client to get the auth cookie
        into the headers.
    """

    async def post(self):
        """ handle login post """
        username = self.get_argument("username", None)
        password = self.get_argument("password", None)
        if username is None or password is None:
            raise HTTPError(403, "username or password is None")

        user = self.login(username, password)  # pylint: disable=E1101
        if asyncio.iscoroutine(user):
            user = await user
        if user:
            self.set_current_user(user)
            self.redirect(self.get_argument("next", "/"))
        else:
            self.set_status(403)
            self.finish("<html><body>Username or password incorrect</body></html>")



"""
    A dict that allows for object-like property access syntax.
"""


class DotDict(dict):
    # because square brackets are too many characters?
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)



"""
    LDAP AUTH
"""
import logging
from tornado.options import options
from ldap3 import Server, Connection, SIMPLE, SYNC, SUBTREE, ALL
from ldap3.core.exceptions import LDAPBindError
from bluemax.web import json_utils
from .login_handler import LoginHandler

AUTH_LDAP_ATTRS = [
    "cn",
    "cn",
    "department",
    "description",
    "displayname",
    "distinguishedname",
    "givenname",  # firstname
    "mail",
    "memberof",
    "name",  # full name
    "objectclass",
    "samaccountname",
    "sn",  # last name or surname
    "telephonenumber",
    "title",
    "useraccountcontrol",
    "userprincipalname",
    "whencreated",
]
AUTH_LDAP_SUPERUSERS = [
    "pshingavi@directbuy.com",
    "sashford@directbuy.com",
    "patrick.lowe@directbuy.com",
    "fleng@directbuy.com",
    "peter.bunyan@directbuy.com",
]


def _login_(username, password):
    try:
        s = Server(options.LDAP_HOST, get_info=ALL)
        c = Connection(
            s,
            auto_bind=True,
            client_strategy=SYNC,
            user=username.strip(),
            password=password.strip(),
            authentication=SIMPLE,
            check_names=True,
        )
        filter_str = f"(&(!(useraccountcontrol=514))(userprincipalname={c.user}))"
        logging.info(filter_str)
        r = c.search(
            search_base="dc=ucctops,dc=com",
            search_filter=filter_str,
            search_scope=SUBTREE,
            attributes=AUTH_LDAP_ATTRS,
        )
        if r is True:
            user = json_utils.loads(c.entries[0].entry_to_json())
            result = {"username": username, "user": user}
            logging.info("logged in %r", result)
            return result
        result = {"username": username}
        logging.info("logged in %r", result)
        return result
    except LDAPBindError as ex:
        logging.info("login failed: %s", ex)


class LDAP3LoginHandler(LoginHandler):  # pylint: disable=W0223
    """ simple pass through handler """

    def login(self, username, password):
        """ if you in we'll auth """
        result = None
        if username in AUTH_LDAP_SUPERUSERS:
            result = _login_(username, password)
        return result



from setuptools import setup, find_packages

try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

    class bdist_wheel(_bdist_wheel):
        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            self.root_is_pure = True


except ImportError:
    bdist_wheel = None


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="bluemax",
    version="0.1.6",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/blueshed/bluemax/",
    packages=find_packages(exclude=["tests.*", "tests"]),
    package_data={"bluemax.web": ["static/*"]},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["tornado", "invoke", "pyyaml"],
    extras_require={
        "sa": ["alembic"],
        "redis": ["aioredis"],
        "all": ["alembic", "aioredis"],
        "dev": [
            "alembic",
            "aioredis",
            "pytest-tornado",
            "wheel",
            "twine",
            "bumpversion",
            "pytest-tornasync",
            "pytest-cov",
            "pytest-docker-fixtures",
            "lovely-pytest-docker",
            "pytest-benchmark",
            "flake8",
            "ldap3",
        ],
    },
    entry_points={"console_scripts": ["bluemax=bluemax.tasks:program.run"]},
)



# pylint: disable=W0621
"""
    Task for the running of bluemax
    The load order:
        if module:
            load module
            load settings
            extend setting
            load logging
            run target
        else
            load settings
            load module
            extend settings
            load logging
            run target
"""
import logging
import yaml
from invoke import task
from bluemax import settings
from .utils import get_module, pid_file, extend_logging

LOGGER = logging.getLogger(__name__)


def bring_up(module=None, config=None):
    """ the standard bring up order """
    if config:
        settings.init_settings(config)
    if module is None:
        module = settings.SETTINGS["procedures"]
    assert module, "module required"
    mod = get_module(module)
    if mod:
        settings.SETTINGS.update(mod)
    extend_settings = settings.get_settings("settings_extend")
    if extend_settings:
        settings.extend(extend_settings, settings.SETTINGS)
    extend_logging(settings.SETTINGS)


@task
def config(_, module=None, config=None):
    """ displays resultant config """
    bring_up(module, config)
    LOGGER.info("\n---\n%s", yaml.dump(settings.SETTINGS))


@task(
    default=True,
    help={
        "pid": "/path/to/pid/file",
        "module": "module to remote",
        "services": "services module",
        "config": "path to config.yml",
    },
)
def server(_, module=None, pid="server.pid", services=None, config=None):
    """ runs a bluemax server with optional services """
    from bluemax.web import server as m_server

    bring_up(module, config)
    if services:
        settings.SETTINGS["services"] = services
    with pid_file(pid):
        m_server.main()


@task(
    help={
        "pid": "/path/to/pid/file",
        "module": "module to remote",
        "services": "services module",
        "config": "path to config.yml",
    }
)
def worker(_, module=None, pid="worker.pid", config=None):
    """ runs a bluemax worker """
    from bluemax.work import worker as m_worker

    bring_up(module, config)
    with pid_file(pid):
        m_worker.main()


@task(
    help={
        "pid": "/path/to/pid/file",
        "module": "services module",
        "config": "path to config.yml",
    }
)
def services(_, module=None, pid="services.pid", services=None, config=None):
    """ runs a bluemax services """
    from bluemax.services import server as s_worker

    bring_up(module, config)
    if services:
        settings.SETTINGS["services"] = services
    with pid_file(pid):
        s_worker.main()



# pylint: disable=C0103
"""
    Like aws_pubsub but for redis
"""
import asyncio
import functools
import logging
from . import rd_session

LOGGER = logging.getLogger(__name__)


class redis_publish:
    """
        decorator to publish to redis

        usage:
        @redis_publish
        def foo():
            # will publish 'foo' to redis:foo
            return 'foo'

        @redis_publish(queue_name='bar')
        def foo():
            # will publish 'foo' to redis:bar
            return 'foo'
    """

    def __init__(self, method=None, key=None, cmd="publish"):
        LOGGER.debug("__init__(%r, %r)", method, key)
        if method and hasattr(method, "__call__"):
            self.method = method
        else:
            self.method = None
        self.key = key
        self.cmd = cmd

    def __get__(self, obj, type_=None):
        return functools.partial(self, obj)

    async def _publish_(self, *args, **kwargs):
        """ Called each time we are called """
        if asyncio.iscoroutinefunction(self.method):
            document = await self.method(*args, **kwargs)
        else:
            document = self.method(*args, **kwargs)
        pool = await rd_session.pool()
        await pool.execute(self.cmd, self.key, document)
        LOGGER.debug("queued %s -> %r", self.key, document)

    def __call__(self, *args, **kwargs):
        """ Called by actor, we return a task """
        LOGGER.debug("__call__(%r, %r) %r", args, kwargs, self.method)
        if self.method is None:
            self.method = args[0]
            return self
        if self.key is None:
            self.key = self.method.__name__

        LOGGER.info("method %r", self.method)
        task = asyncio.ensure_future(self._publish_(*args, **kwargs))
        return task


class redis_subscribe:
    """
        decorator subscribe to redis

        usage:
        @redis_subscribe
        def foo(message):
            # will subscribe to redis:foo
            print(message)

        @redis_subscribe(queue_name='bar')
        def foo(message):
            # will subscribe to redis:bar
            print(message)
    """

    def __init__(self, method=None, key=None):
        LOGGER.debug("__init__(%r, %r)", method, key)
        if method and hasattr(method, "__call__"):
            self.method = method
        else:
            self.method = None
        self.key = key

    def __get__(self, obj, type_=None):
        return functools.partial(self, obj)

    async def _subscribe_(self, *args, **kwargs):
        """ read our queue and call our method """
        conn = await rd_session.connect()
        response = await conn.subscribe(self.key)
        channel = response[0]
        LOGGER.info("subscribed to: %s", self.key)
        try:
            while await channel.wait_message():
                document = await channel.get(encoding="utf-8")
                LOGGER.debug("Got Message: %s", document)
                _args = list(args) + [document]
                if asyncio.iscoroutinefunction(self.method):
                    asyncio.ensure_future(self.method(*_args, **kwargs))
                else:
                    self.method(*_args, **kwargs)
        finally:
            conn.close()

    def __call__(self, *args, **kwargs):
        """ called to set us running """
        LOGGER.debug("__call__(%r, %r) %r", args, kwargs, self.method)
        if self.method is None:
            self.method = args[0]
            return self
        if self.key is None:
            self.key = self.method.__name__
        task = asyncio.ensure_future(self._subscribe_(*args, **kwargs))
        return task


class redis_cmd:
    """
        decorator subscribe to redis

        usage:
        @redis_subscribe
        def foo(message):
            # will subscribe to redis:foo
            print(message)

        @redis_subscribe(queue_name='bar')
        def foo(message):
            # will subscribe to redis:bar
            print(message)
    """

    def __init__(self, method=None, key=None, cmd="get", cmd_args=None):
        LOGGER.debug("__init__(%r, %r)", method, key)
        if method and hasattr(method, "__call__"):
            self.method = method
        else:
            self.method = None
        self.key = key
        self.cmd = cmd
        self.cmd_args = cmd_args if cmd_args else []
        self.conn = None

    def __get__(self, obj, type_=None):
        return functools.partial(self, obj)

    async def execute(self, cmd, key):
        """ raw execute of cmd on redis connection """
        result = await self.conn.execute(cmd, key, *self.cmd_args)
        LOGGER.debug("Result %s", result)
        return result

    async def _subscribe_(self, *args, **kwargs):
        """ read our queue and call our method """
        self.conn = await rd_session.pool()
        LOGGER.info("cmd %s to: %s", self.cmd, self.key)
        while True:
            document = await self.execute(self.cmd, self.key)
            LOGGER.debug("Got %s: %s", self.key, document)
            _args = list(args) + [document]
            if asyncio.iscoroutinefunction(self.method):
                asyncio.ensure_future(self.method(*_args, **kwargs))
            else:
                self.method(*_args, **kwargs)

    def __call__(self, *args, **kwargs):
        """ called to set us running """
        LOGGER.debug("__call__(%r, %r) %r", args, kwargs, self.method)
        if self.method is None:
            LOGGER.info("method null: %s", args)
            self.method = args[0]
            return self
        if self.key is None:
            self.key = self.method.__name__
        LOGGER.info("method %r", self.method)
        task = asyncio.ensure_future(self._subscribe_(*args, **kwargs))
        return task



# pylint: disable=C0103
"""
    Classes that wrap functions to publish to sns or consume sqs
"""
import asyncio
import inspect
import functools
import logging
import aiobotocore
from bluemax.web.json_utils import dumps, loads
from bluemax.settings import get_settings

LOGGER = logging.getLogger(__name__)


class AwsCache:
    """
        A cache for AWS clients and Sessions
    """

    clients = {}
    session = None
    region = None
    account_id = None

    @classmethod
    def get_endpoint_url(cls, kind):
        """ Returns endpoint url overide from settings """
        endpoint_urls = get_settings("endpoints", {})
        return endpoint_urls.get(kind)

    @classmethod
    def client(cls, kind, endpoint_url=None):
        """ Returns a cached client if available """
        client = cls.clients.get(kind)
        if client is None:
            if cls.session is None:
                loop = asyncio.get_event_loop()
                cls.session = aiobotocore.get_session(loop=loop)
            if endpoint_url is None:
                endpoint_url = cls.get_endpoint_url(kind)
            client = cls.session.create_client(
                kind, endpoint_url=endpoint_url
            )
            cls.clients[kind] = client
        return client

    @classmethod
    async def get_region_account_id(cls):
        """ returns the region and account of current credentials """
        if cls.region is None or cls.account_id is None:
            if get_settings("endpoints"):
                return "us-east-1", "123456789012"
            client = cls.client("sts")
            cls.region = client.meta.region_name
            response = await client.get_caller_identity()
            cls.account_id = response["Account"]
        return cls.region, cls.account_id

    @classmethod
    async def get_topic_arn(cls, topic):
        """
            Compose an arn from session data
        """
        region, account_id = await cls.get_region_account_id()
        topic_arn = f"arn:aws:sns:{region}:{account_id}:{topic}"
        return topic_arn

    @classmethod
    async def get_queue_url(cls, queue: str):
        """
            Returns the queue_url for queue_name
        """
        endpoint_url = cls.get_endpoint_url("sqs")
        if endpoint_url:
            # queue_url = f"{endpoint_urls['sqs']}/queue/{queue}"
            queue_url = f"http://sqs:9324/queue/{queue}"
        else:
            client = cls.client("sts")
            region = client.meta.region_name
            response = await client.get_caller_identity()
            account_id = response["Account"]
            queue_url = (
                f"https://{region}.queue.amazonaws.com/{account_id}/{queue}"
            )
        return queue_url

    @classmethod
    async def apublish(cls, topic_name, message):
        """ publishes a message to a topic """
        message = dumps(message)
        stage = get_settings("STAGE")
        topic_arn = f"{topic_name}-{stage}"
        topic_arn = await cls.get_topic_arn(topic_arn)
        return await AwsCache.client("sns").publish(TopicArn=topic_arn, Message=message)

    @classmethod
    async def apost(cls, queue_name, message):
        """ posts a message into a queue """
        message = dumps({
            "Message": dumps(message)
        })
        stage = get_settings("STAGE")
        queue_url = f"{queue_name}-{stage}"
        queue_url = await cls.get_queue_url(queue_url)
        return await AwsCache.client("sqs").send_message(QueueUrl=queue_url, MessageBody=message)


def publish(topic_name, message):
    """ returns a future from apublish """
    return asyncio.ensure_future(AwsCache.apublish(topic_name, message))


def post(queue_name, message):
    """ returns a future for apost """
    return asyncio.ensure_future(AwsCache.apost(queue_name, message))


class aws_publish:
    """
        decorator to publish to aws topic

        usage:
        @aws_publish
        def foo():
            # will publish 'foo' to sns:foo
            return 'foo'

        @aws_publish(topic_name='bar')
        def foo():
            # will publish 'foo' to sns:bar
            return 'foo'
    """

    TOPICS = []

    def __init__(self, method=None, topic_name=None):
        LOGGER.debug("__init__(%r, %r)", method, topic_name)
        if method and hasattr(method, "__call__"):
            self.method = method
        else:
            self.method = None
        self.topic_name = topic_name
        self.topic_arn = None

    def __get__(self, obj, type_=None):
        return functools.partial(self, obj)

    async def _publish_(self, *args, **kwargs):
        if inspect.iscoroutinefunction(self.method):
            document = await self.method(*args, **kwargs)
        else:
            document = self.method(*args, **kwargs)
        LOGGER.debug("published %s -> %r", self.topic_name, document)
        if self.topic_arn is None:
            stage = get_settings("STAGE")
            topic_arn = f"{self.topic_name}-{stage}"
            self.topic_arn = await AwsCache.get_topic_arn(topic_arn)
        message = dumps(document)
        LOGGER.info("publishing: %s -> %r", self.topic_arn, message)
        response = await AwsCache.client("sns").publish(
            TopicArn=self.topic_arn, Message=message
        )
        print(response)

    def __call__(self, *args, **kwargs):
        LOGGER.debug("__call__(%r, %r) %r", args, kwargs, self.method)
        if self.method is None:
            self.method = args[0]
            return self
        if self.topic_name is None:
            self.topic_name = self.method.__name__
        asyncio.ensure_future(self._publish_(*args, **kwargs))
        return None


class aws_subscribe:
    """
        decorator subscribe to aws queue

        usage:
        @aws_subscribe
        def foo(message):
            # will subscribe to sqs:foo
            print(message)

        @aws_subscribe(queue_name='bar')
        def foo(message):
            # will subscribe to sqs:bar
            print(message)
    """

    QUEUES = []

    def __init__(self, method=None, queue_name=None):
        LOGGER.debug("__init__(%r, %r)", method, queue_name)
        if method and hasattr(method, "__call__"):
            self.method = method
        else:
            self.method = None
        self.queue_name = queue_name

    def __get__(self, obj, type_=None):
        return functools.partial(self, obj)

    async def _subscribe_(self):
        stage = get_settings("STAGE")
        sqs_client = AwsCache.client("sqs")
        queue_url = await AwsCache.get_queue_url(f"{self.queue_name}-{stage}")
        LOGGER.debug("subscribed to: %s", queue_url)
        while True:
            response = await sqs_client.receive_message(QueueUrl=queue_url)
            if "Messages" in response:
                for msg in response["Messages"]:
                    LOGGER.info("Got msg %r", msg["Body"])
                    message = loads(msg["Body"])
                    document = loads(message["Message"])
                    if inspect.iscoroutinefunction(self.method):
                        asyncio.ensure_future(self.method(document))
                    else:
                        self.method(document)
                    # Need to remove msg from queue or else it'll reappear
                    await sqs_client.delete_message(
                        QueueUrl=queue_url, ReceiptHandle=msg["ReceiptHandle"]
                    )
            else:
                LOGGER.debug("No messages in queue")

    def __call__(self, *args, **kwargs):
        LOGGER.debug("__call__(%r, %r) %r", args, kwargs, self.method)
        if self.method is None:
            self.method = args[0]
            return self
        if self.queue_name is None:
            self.queue_name = self.method.__name__
        asyncio.ensure_future(self._subscribe_(*args, **kwargs))
        return "subscribed"
